/**
 * Copyright (c) 2009-2010, Arnoud Philip
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package com.dexels.blog;

import java.awt.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.io.File;
import java.util.ArrayList;

import javax.imageio.ImageIO;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;

import org.pushingpixels.trident.Timeline;
import org.pushingpixels.trident.Timeline.RepeatBehavior;
import org.pushingpixels.trident.Timeline.TimelineState;
import org.pushingpixels.trident.callback.TimelineCallback;
import org.pushingpixels.trident.callback.TimelineCallbackAdapter;
import org.pushingpixels.trident.ease.Sine;
import org.pushingpixels.trident.ease.Spline;
import org.pushingpixels.trident.ease.TimelineEase;
import org.pushingpixels.trident.interpolator.PropertyInterpolator;


public class PeelEffectLayer extends JPanel {

	private final double	peelDistanceThreshold = 100.0;
	private Point 			peelLocation = new Point(1, 1);
	private Point 			animationOrigin;
	private Point			flipOverTarget;
	private Point 			animTarget;
	private BufferedImage 	foregroundImage, backgroundImage, back, currentFront, currentBack;
	private ArrayList<PeelListener> listeners = new ArrayList<PeelListener>();
//	private Color 			backSideColor = Color.decode("#97CCC2"); //50da26
	private Color 			backSideColor = Color.decode("#50da26");
	private BufferedImage 	backSideImage = null;
	private Rectangle 		prevBounds;
	private boolean 		flipOver = false;
	private boolean 		isPeeling = false;
	private double 			mouseDistance = 0.0;
	private int 			dragDir = 0;
	private Point 			locFrom = new Point(0, 0);
	private BufferedImage 	clippedImage;

	public boolean 			isDragging = false;
	public  boolean 		isAnimating = false;
	public  double 			peelFactor = 0.0;	
	

	// Start with right bottom
	public PeelEffectLayer() {

		addMouseMotionListener(new MouseMotionListener() {

			public void mouseDragged(MouseEvent e) {
				peelLocation = e.getPoint();							
				repaint();
			}

			/*
			 * Determine which corner the mouse pointer is closest to
			 * if the mouse pointer's distance from that corner is 
			 * less then the peelDistanceThreshold parameter then animate
			 * the peel effect towards the mouse pointer. If not, animate back.
			 */
			public void mouseMoved(MouseEvent e) {
				if (isAnimating) {
					return;
				}
				
				// Determine corner
				Point cornerLocation = new Point(0, 0);
				Rectangle bounds = getBounds();
				if (e.getPoint().y < bounds.height / 2) {
					cornerLocation.y = bounds.y;
				} else {
					cornerLocation.y = bounds.y + bounds.height;
				}
				if (e.getPoint().x < bounds.width / 2) {
					cornerLocation.x = bounds.x;
				} else {
					cornerLocation.x = bounds.x + bounds.width;
				}

				double dx = cornerLocation.x - e.getPoint().x;
				double dy = cornerLocation.y - e.getPoint().y;

				// Calculate distance and animate accordingly
				mouseDistance = Math.sqrt((Math.pow(dx, 2.0) + Math.pow(dy, 2.0)));
				if (mouseDistance < peelDistanceThreshold) {
					startPeel(e);
				} else if (peelLocation.x > 1 && peelLocation.y > 1) {
					animateBack();
				}
			}
		});

		/*
		 * Record start and end of dragging
		 */
		addMouseListener(new MouseAdapter() {
			
			// Start when a button is pressed
			public void mousePressed(MouseEvent e) {
				isDragging = true;
			}

			// Release when the button is released
			public void mouseReleased(MouseEvent e) {
				isDragging = false;
				if (!isAnimating) {
					animateBack();
				}
			}
		});

		addMouseMotionListener(new MouseMotionAdapter() {
			public void mouseExited(MouseEvent e) {
				if (isDragging) {					
					isDragging = false;
					peelLocation.x = 1;
					peelLocation.y = 1;
					// repaint();
				}
			}
		});

		setBackground(Color.white);
	}
	
	/*
	 * Register listeners
	 */
	public void addPeelListener(PeelListener l){
		if(!listeners.contains(l)){
			listeners.add(l);
		}
	}
	
	/*
	 * Notify listeners on the termination of a peel
	 */
	private void firePeelComplete(boolean full_peel){
		for(int i=0;i<listeners.size();i++){
			listeners.get(i).peelStopped(full_peel);
		}
	}
	
	/*
	 *  Getter for the current peel location
	 */
	public Point getPeelLoc(){
		return peelLocation;
	}
	
	/* 
	 * Set the color that is displayed at the backside of the page.
	 * This can be used instead of an image.
	 */
	public void setBackSideColor(Color c){
		this.backSideColor = c;
	}
	
	/*
	 *  Set the image that is displayed as the backside of the page.
	 *  This can be used instead of a color.
	 */
	public void setBackSideImage(BufferedImage img){
		this.backSideImage = img;
	}		
	
	/* 
	 * Set current page and next page comopnents
	 * These components will be converted into buffered images and
	 * used to draw the peel effect. C1 is the current page, C2 is
	 * the next page that will be visible after a page flip.
	 */
	public void setComponents(JComponent c1, JComponent c2) {
		clippedImage = null;

		BufferedImage b1 = new BufferedImage(c1.getWidth(), c1.getHeight(), BufferedImage.TYPE_INT_RGB);
		BufferedImage b2 = new BufferedImage(c2.getWidth(), c2.getHeight(), BufferedImage.TYPE_INT_RGB);
		
		Graphics2D b1g = b1.createGraphics();
		Graphics2D b2g = b2.createGraphics();		

		back = new BufferedImage(c2.getWidth(), c2.getHeight(), BufferedImage.TYPE_INT_RGB);;
		Graphics2D backG = back.createGraphics();
		if(backSideColor != null){
			backG.setColor(backSideColor);
		}else{
			backG.setColor(Color.black);
		}
		backG.fillRect(0, 0, c2.getWidth(), c2.getHeight());
		if(backSideImage != null){
			int x = c2.getWidth()/2 - backSideImage.getWidth()/2;
			int y = c2.getHeight()/2 - backSideImage.getHeight()/2;
			backG.drawImage(backSideImage, x, y, backSideImage.getWidth(), backSideImage.getHeight(), null);
		}		
		
		c1.print(b1g);		
		b1g.dispose();
		currentFront = b1;
		foregroundImage = currentFront;

		c1.setVisible(false);
		c2.setVisible(true);
		c2.print(b2g);
		b2g.dispose();
		currentBack = b2;
		backgroundImage = currentBack;

		c1.setVisible(true);
		c2.setVisible(false);
	}

	/*
	 *  Set up variables to start the peel effect
	 */
	private void startPeel(MouseEvent e) {
		peelLocation = e.getPoint();		
		Rectangle bounds = getBounds();
		if (!isDragging) {
			if (peelLocation.y < bounds.height / 2) {
				locFrom.y = bounds.y;
			} else {
				locFrom.y = bounds.y + bounds.height;
			}
			if (peelLocation.x < bounds.width / 2) {
				locFrom.x = bounds.x;
			} else {
				locFrom.x = bounds.x + bounds.width;
			}
			determineDragDirection(locFrom);
			peelFactor = 0.0;
		}
		if(!isPeeling){
			aniticipate();
		}
		isPeeling = true;
		if(!isAnimating){
			repaint();
		}
	}
	
	/* 
	 * Quickly animate the tip of the paper towards the mousepointer
	 */
	private void aniticipate(){
		animTarget = new Point(peelLocation.x, peelLocation.y);
		int duration = 75;
		animationOrigin = new Point(locFrom.x, locFrom.y);
		
		isAnimating = true;
		
		Timeline timeline = new Timeline(this);		
		timeline.addPropertyToInterpolate("peelFactor", 0.0f, 1.0f, new FloatInterpolator());
		timeline.setDuration(duration);
		timeline.setEase(new Spline(.4f));
		
		timeline.addCallback(new TimelineCallbackAdapter(){
			public void onTimelineStateChanged(TimelineState prevState, TimelineState newState, float a, float b) {
				if(newState == TimelineState.IDLE){
					isAnimating = false;					
				}				
			}			
		});
		timeline.play();
	}
	
	/*
	 *  Set up variables to start the peel effect
	 *  Base of a point rather then a mouseevent
	 */
	public void setPeelPoint(Point peelLocation) {
		this.peelLocation = peelLocation;
	}
	
	/*
	 *  Determine whether we are dragging the mouse in the right direction
	 */
	private boolean isLocationValid(int dragDir, Point location){
		switch (dragDir) {
		case 0:
			// Top left
			if(location.x < 0 || location.y < 0){
				return false;
			}
			break;
		case 1:
			// Top right
			if(location.x > getWidth() || location.y < 0){
				return false;
			}
			break;
		case 2:
			if(location.x > getWidth() || location.y > getHeight()){
				return false;
			}
			break;
		case 3:
			if(location.x < 0 || location.y > getHeight()){
				return false;
			}			
			break;
		}
		
		return true;

	}

	/* 
	 * Determine the direction of the drag
	 */
	private void determineDragDirection(Point loc) {
		Rectangle bounds = getBounds();
		if (loc.x == 0 && loc.y == 0) {
			dragDir = 0;
		} else if (loc.x == bounds.width && loc.y == 0) {
			dragDir = 1;
		} else if (loc.x == bounds.width && loc.y == bounds.height) {
			dragDir = 2;
		} else if (loc.x == 0 && loc.y == bounds.height) {
			dragDir = 3;
		}
	}


	/*
	 *  Custom paint routines to display the peel effect. 
	 */
	public void paint(Graphics g) {
		Rectangle bnds = getBounds();

		// Resize container
		// Store the current page in a buffered image, clippedImage if the size of the container changed.
		if (!bnds.equals(prevBounds) || clippedImage == null) {
			prevBounds = bnds;
			clippedImage = new BufferedImage(bnds.width, bnds.height, BufferedImage.TYPE_INT_ARGB);
			Graphics2D clippedG = clippedImage.createGraphics();
			clippedG.drawImage(currentFront, 0, 0, currentFront.getWidth(), currentFront.getHeight(), null);
			peelLocation.x = 1;
			peelLocation.y = 1;
			clippedG.dispose();
		}

		Graphics2D g2 = (Graphics2D) g.create();
		// If we are peeling, continue
		if (peelLocation != null) {
		
			// Fill background with black.
			g2.setColor(Color.black);
			g2.fill(getBounds());
			
			double dx = locFrom.x - peelLocation.x;
			double dy = locFrom.y - peelLocation.y;
			double rc = dy / dx;
			// Find B
			double b = -rc * peelLocation.x + peelLocation.y;
			// y = rc*x + b

			double dpx = peelLocation.x + (locFrom.x - peelLocation.x) / 2.0;
			double dpy = peelLocation.y + (locFrom.y - peelLocation.y) / 2.0;

			double b_orth = (1.0 / rc) * dpx + dpy;

			// if we know Y what is x?
			int cx = (int) ((locFrom.y - b_orth) / (-1.0 / rc));
			int o2 = (int) ((-1.0 / rc) * locFrom.x + b_orth);

			Point p1 = new Point(cx, locFrom.y);
			Point p2 = new Point(locFrom.x, o2);

			// Now determine our shape(s)

			Polygon p = new Polygon(new int[] { peelLocation.x, p1.x, p2.x }, new int[] { peelLocation.y, p1.y, p2.y }, 3);
			Polygon q = new Polygon(new int[] { p2.x, p1.x, locFrom.x }, new int[] { p2.y, p1.y, locFrom.y }, 3);

			// Main Container
			g2.drawImage(clippedImage, 0, 0, bnds.width, bnds.height, null);

			// TODO, check if peellocation is in legal territory..
			
			
			if (((int) dpx != peelLocation.x) && ((int) dpy != peelLocation.y) && isPeeling && isLocationValid(dragDir, peelLocation)) {
				LinearGradientPaint gpBounds = new LinearGradientPaint((float) peelLocation.x, (float) peelLocation.y, (float) dpx, (float) dpy, new float[] { 0.0f, 1.0f }, new Color[] { new Color(1.0f, 1.0f, 1.0f, 0.0f), new Color(0.0f, 0.0f, 0.0f, 0.5f) });
				g2.setPaint(gpBounds);
				g2.fill(bnds);
			}

			// Peeled
			// Fill with rotated bgimage if available

			if (isPeeling && isLocationValid(dragDir, peelLocation)) {

				double delta = dx / dy; // tan(theta) = delta
				double theta = Math.atan(delta);

				// Find out translation
				int tx = 0;
				int ty = 0;
				Point rotator = new Point(0, 0);

				switch (dragDir) {
				case 0:
					tx = -bnds.width + peelLocation.x;
					ty = peelLocation.y;
					rotator.x = bnds.width;
					break;
				case 1:
					tx = peelLocation.x;
					ty = peelLocation.y;
					break;
				case 2:
					tx = peelLocation.x;
					ty = peelLocation.y - bnds.height;
					rotator.x = 0;
					rotator.y = bnds.height;
					break;
				case 3:
					tx = -bnds.width + peelLocation.x;
					ty = peelLocation.y - bnds.height;
					rotator.x = bnds.width;
					rotator.y = bnds.height;
					break;
				}

				
				Shape clip = g2.getClip();
				g2.setClip(p);
				g2.translate(tx, ty);
				g2.rotate(-2 * theta + Math.PI, rotator.x, rotator.y);

				g2.drawImage(back, 0, 0, back.getWidth(), back.getHeight(), null);

				g2.rotate(2 * theta + Math.PI, rotator.x, rotator.y);
				g2.translate(-tx, -ty);
				g2.setClip(clip);

				if (((int) dpx != peelLocation.x) && ((int) dpy != peelLocation.y)) {
					LinearGradientPaint gpPeel = new LinearGradientPaint((float) peelLocation.x, (float) peelLocation.y, (float) dpx, (float) dpy, new float[] { 0.0f, 0.7f, 0.99f, 1.0f }, new Color[] { new Color(0.0f, 0.0f, 0.0f, 0.1f), new Color(0.0f, 0.0f, 0.0f, 0.2f), new Color(1.0f, 1.0f, 1.0f, 0.5f), new Color(0.0f, 0.0f, 0.0f, 0.5f) });
					g2.setPaint(gpPeel);
					g2.fill(p);
				}

				// New page
				g2.setColor(Color.black);
				clip = g2.getClip();
				g2.setClip(q);

				g2.drawImage(currentBack, 0, 0, currentBack.getWidth(), currentBack.getHeight(), null);
				
				if (((int) dpx != peelLocation.x) && ((int) dpy != peelLocation.y)) {
					LinearGradientPaint gpPeel = new LinearGradientPaint((float) dpx, (float) dpy, locFrom.x, locFrom.y, new float[] { 0.0f, 0.25f}, new Color[] { new Color(0.0f, 0.0f, 0.0f, 0.6f), new Color(0.0f, 0.0f, 0.0f, 0.0f)});
					g2.setPaint(gpPeel);
					g2.fill(q);
				}
				g2.setClip(clip);
				g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
				g2.setColor(new Color(0.2f, 0.2f, 0.2f, 1f));
				g2.drawLine(p1.x, p1.y, p2.x, p2.y);	
				g2.setColor(new Color(1f, 1f, 1f, 1f));
				g2.drawLine(p1.x, p1.y, peelLocation.x, peelLocation.y);
				g2.drawLine(p2.x, p2.y, peelLocation.x, peelLocation.y);
				g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
			}
			g2.dispose();
		}
	}

	/*
	 * Animate to destination. 
	 * This can either be a flip over or a reset to the origin.
	 */
	public void animateBack() {
		animationOrigin = new Point(peelLocation.x, peelLocation.y);
		determineFlipOver(animationOrigin);
		int duration = 250;
		if (flipOver) {
			determineFlipOverTarget(animationOrigin);
			animTarget = new Point(flipOverTarget.x, flipOverTarget.y);
//			duration = 500;
		} else {
			animTarget = new Point(locFrom.x, locFrom.y);
		}
		isAnimating = true;
		
		Timeline timeline = new Timeline(this);		
		timeline.addPropertyToInterpolate("peelFactor", 0.0f, 1.0f, new FloatInterpolator());
		timeline.setDuration(duration);
		timeline.setEase(new Spline(.4f));
		
		timeline.addCallback(new TimelineCallbackAdapter(){
			public void onTimelineStateChanged(TimelineState prevState, TimelineState newState, float a, float b) {
				if(newState == TimelineState.IDLE){
					isAnimating = false;
					isPeeling = false;					
					if (flipOver) {
						try {
							repaint();
							firePeelComplete(true);
						} catch (Exception e) {
							e.printStackTrace();
						}
					}else{
						firePeelComplete(false);
					}
				}				
			}			
		});
		timeline.play();
	}

	/*
	 *  Setter used by the animation timeline to indicate how far the aniumation has progressed.
	 */
	public void setPeelFactor(double d) {
		this.peelFactor = d;
		double x = (animTarget.x - animationOrigin.x) * d + animationOrigin.x;
		double y = (animTarget.y - animationOrigin.y) * d + animationOrigin.y;
		peelLocation.x = (int) x;
		peelLocation.y = (int) y;
		repaint();
	}

	/*
	 *  Determine if the user has dragged far enough to cause a page flip.
	 */
	private void determineFlipOver(Point origin) {
		Rectangle bounds = getBounds();
		flipOver = false;
		if (origin.x < bounds.width / 2 && (dragDir == 1 || dragDir == 2)) {
			flipOver = true;
		}
		if (origin.x > bounds.width / 2 && (dragDir == 0 || dragDir == 3)) {
			flipOver = true;
		}
		if (origin.y < bounds.height / 2 && (dragDir == 2 || dragDir == 3)) {
			flipOver = true;
		}
		if (origin.y > bounds.height / 2 && (dragDir == 0 || dragDir == 1)) {
			flipOver = true;
		}
		flipOver = flipOver && isPeeling;
	}

	/*
	 * Determine the target location of the page flip animation.
	 */
	private void determineFlipOverTarget(Point origin) {
		// determine quadrant of origin
		Rectangle bounds = getBounds();
		switch (dragDir) {
		case 0:
			if (origin.x > bounds.width / 2) { // FlipTarget is top-right
				flipOverTarget = new Point(2 * bounds.width, 0);
			} else { // FlipTarget is bottom-left
				flipOverTarget = new Point(0, 2 * bounds.height);
			}
			break;
		case 1:
			if (origin.x < bounds.width / 2) { // FlipTarget is top-left
				flipOverTarget = new Point(-bounds.width, 0);
			} else { // FlipTarget is bottom-right
				flipOverTarget = new Point(bounds.width, 2 * bounds.height);
			}
			break;
		case 2:
			if (origin.x < bounds.width / 2) { // FlipTarget is bottom-left
				flipOverTarget = new Point(-bounds.width, bounds.height);
			} else { // FlipTarget is top-right
				flipOverTarget = new Point(bounds.width, -bounds.height);
			}
			break;
		case 3:
			if (origin.x > bounds.width / 2) { // FlipTarget is bottom-right
				flipOverTarget = new Point(2 * bounds.width, bounds.height);
			} else { // FlipTarget is top-left
				flipOverTarget = new Point(0, -bounds.height);
			}
			break;
		}
	}
	
	/*
	 * Utility class to interpolate floats
	 */
	class FloatInterpolator implements PropertyInterpolator<Float>{

		@Override
		public Class getBasePropertyClass() {
			return Float.class;
		}

		@Override
		public Float interpolate(Float from, Float to, float pos) {
			return pos * to;
		}
		
	}
}
