/**
 * Copyright (c) 2009-2010, Arnoud Philip
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package com.dexels.blog;

import java.awt.*;
import java.awt.event.FocusEvent;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.ArrayList;

import javax.imageio.ImageIO;
import javax.swing.*;

import org.jdesktop.jxlayer.JXLayer;
import org.jdesktop.jxlayer.plaf.AbstractLayerUI;

public class PeelPanel extends JPanel implements PeelListener {
	private final static int PEEL_TOP_LEFT = 0;
	private final static int PEEL_TOP_RIGHT = 1;
	private final static int PEEL_BOTTOM_RIGHT = 2;
	private final static int PEEL_BOTTOM_LEFT = 3;
	private final ArrayList<Integer> enabledCorners = new ArrayList<Integer>();
	private final ArrayList<JComponent> components = new ArrayList<JComponent>();
	private static final long serialVersionUID = 1L;
	private final JPanel main = new JPanel();
	private final JXLayer<JComponent> layer;
	private final PeelEffectLayer effectLayer = new PeelEffectLayer();

	private int browseDirection = 1;
	private JComponent visibleComponent = null;
	private JComponent next = null;
	private double mouseDistance = 0.0;
	private double max_mouseDistance = 50.0;

	public PeelPanel() {
		effectLayer.addPeelListener(this);
		layer = new JXLayer<JComponent>(main);
		setLayout(new BorderLayout());

		final MouseDrawingUI<JComponent> ui = new MouseDrawingUI<JComponent>() {
		
			@Override
			protected void processMouseEvent(MouseEvent e, JXLayer<? extends JComponent> l) {
				super.processMouseEvent(e, l);
				if(e.getID() == MouseEvent.MOUSE_EXITED){	
					Component c = e.getComponent();
					Rectangle bounds = c.getBounds();
					Point p = SwingUtilities.convertPoint(e.getComponent(), e.getPoint(), l);
					if(!bounds.contains(p) && !effectLayer.isDragging){
						//effectLayer.animateBack();
						layer.setView(visibleComponent);
					}
				}
			}



			private static final long serialVersionUID = 1L;

			
			
			@Override
			protected void processMouseMotionEvent(MouseEvent e, JXLayer<? extends JComponent> l) {
				super.processMouseMotionEvent(e, l);				
				if (effectLayer.isAnimating) {
					return;
				}
				Point mouseLocation = SwingUtilities.convertPoint(e.getComponent(), e.getPoint(), l);

				boolean valid = true;
				Point cornerLocation = new Point(0, 0);
				Rectangle bounds = l.getBounds();

				if (mouseLocation.y < bounds.height / 2) {
					cornerLocation.y = bounds.y;
				} else {
					cornerLocation.y = bounds.y + bounds.height;
				}
				if (mouseLocation.x < bounds.width / 2) {
					cornerLocation.x = bounds.x;
					browseDirection = -1;
				} else {
					cornerLocation.x = bounds.x + bounds.width;
					browseDirection = 1;
				}

				double dx = cornerLocation.x - e.getPoint().x;
				double dy = cornerLocation.y - mouseLocation.y;
				effectLayer.setPeelPoint(e.getPoint());
				mouseDistance = Math.sqrt((Math.pow(dx, 2.0) + Math.pow(dy, 2.0)));
				if (isCornerEnabled(cornerLocation) && mouseDistance < max_mouseDistance) {
					startPeel(e);
				}
			}
		};
		layer.setUI(ui);
		add(layer, BorderLayout.CENTER);

		// By default all corners are enabled for peeling
		enabledCorners.add(PEEL_TOP_LEFT);
		enabledCorners.add(PEEL_TOP_RIGHT);
		enabledCorners.add(PEEL_BOTTOM_RIGHT);
		enabledCorners.add(PEEL_BOTTOM_LEFT);
	}

	public void removeComponent(JComponent c) {
		components.remove(c);
	}

	public void addComponent(JComponent c) {
		components.add(c);
		c.doLayout();
		c.repaint();
		visibleComponent = c;
		main.add(c, BorderLayout.CENTER);
		for (int i = 0; i < components.size(); i++) {
			components.get(i).setVisible(false);
		}
		c.setVisible(true);
		layer.setView(c);
	}

	private void startPeel(MouseEvent e) {
		if (layer.getView() == effectLayer) {
			return;
		}
		layer.setView(effectLayer);

		int idx = components.indexOf(visibleComponent);
		if (idx > -1) {
			int nextComp = idx + browseDirection;
			if (nextComp == -1) {
				nextComp = components.size() - 1;
			} else if (nextComp == components.size()) {
				nextComp = 0;
			}
			next = components.get(nextComp);
			next.setBounds(visibleComponent.getBounds());
			next.doLayout();
			effectLayer.setComponents(visibleComponent, next);
		}
	}

	public void peelStopped(boolean full_peel) {
		for (int i = 0; i < components.size(); i++) {
			components.get(i).setVisible(false);
		}
		if (!full_peel) {
			next.setVisible(false);
			visibleComponent.setVisible(true);
			layer.setView(visibleComponent);
		} else {
			visibleComponent.setVisible(false);
			visibleComponent = next;
			visibleComponent.setVisible(true);
			layer.setView(visibleComponent);
		}
	}

	/*
	 * To do, set specific corners of the screen to peel enabled or not.
	 */
	public void setEnableCorner(int corner, boolean enabled) {
		if (enabled && !enabledCorners.contains(corner)) {
			enabledCorners.add(corner);
		} else if (enabledCorners.contains(corner)) {
			enabledCorners.remove(new Integer(corner));
		}
	}

	private boolean isCornerEnabled(Point cornerLocation) {
		if (cornerLocation.y < getBounds().height / 2) { // top
			if (cornerLocation.x < getBounds().width / 2) { // left
				return enabledCorners.contains(PEEL_TOP_LEFT);
			} else { // right
				return enabledCorners.contains(PEEL_TOP_RIGHT);
			}
		} else { // bottom
			if (cornerLocation.x < getBounds().width / 2) { // left
				return enabledCorners.contains(PEEL_BOTTOM_LEFT);
			} else { // right
				return enabledCorners.contains(PEEL_BOTTOM_RIGHT);
			}
		}
	}

	/*
	 * Set the backside color of all pages
	 */
	public void setBacksideColor(Color backSideColor) {
		effectLayer.setBackSideColor(backSideColor);
	}

	/*
	 * Set the backside image of all pages
	 */
	public void setBacksideImage(BufferedImage backSideImage) {
		effectLayer.setBackSideImage(backSideImage);
	}

	// UI which allows to paint on top of the components, what we don't do here
	public static class MouseDrawingUI<V extends JComponent> extends AbstractLayerUI<V> {
		private static final long serialVersionUID = 1L;

		protected void paintLayer(Graphics2D g2, JXLayer<? extends V> l) {
			super.paintLayer(g2, l);
//			g2.setColor(new Color(0,0,0,50));
//			g2.fill(l.getBounds());
//			Composite old = g2.getComposite();
//			AlphaComposite comp = AlphaComposite.getInstance(AlphaComposite.SRC_ATOP);
//			g2.setComposite(comp);
//			g2.setColor(new Color(255,255,255,150));
//			g2.fillOval(-55,-55,100,100);
		}
		
		
	}

	public static void main(String[] arg) {

		JFrame frame = new JFrame("Peeling panels, by Arnoud Philip 2010");
		PeelPanel peelContainer = new PeelPanel();

		JPanel red = new JPanel();
		JPanel grn = new JPanel();
		JPanel blu = new JPanel();
		JPanel org = new JPanel();
		JPanel cya = new JPanel();
		red.setBackground(Color.red);
		grn.setBackground(Color.green);
		blu.setBackground(Color.blue);
		org.setBackground(Color.orange);
		cya.setBackground(Color.cyan);

		grn.setLayout(new GridBagLayout());
		grn.add(new JTextField("Works great over swing components too!"), new GridBagConstraints(0, 0, 1, 1, 1.0, 0.0, GridBagConstraints.NORTH, GridBagConstraints.HORIZONTAL, new Insets(2, 2, 2, 2), 0, 0));
		JTextArea area = new JTextArea("aaap");
		grn.add(area, new GridBagConstraints(0, 1, 1, 1, 1.0, 1.0, GridBagConstraints.NORTH, GridBagConstraints.BOTH, new Insets(2, 2, 2, 2), 0, 0));

		peelContainer.addComponent(red);
		peelContainer.addComponent(grn);
		peelContainer.addComponent(blu);
		peelContainer.addComponent(org);
		peelContainer.addComponent(cya);

//		try {
//			peelContainer.setBacksideImage(ImageIO.read(PeelPanel.class.getResource("paperweight.png")));
			peelContainer.setBacksideColor(Color.cyan);
//		} catch (IOException e) {
//			e.printStackTrace();
//		}

		frame.getContentPane().setLayout(new BorderLayout());
		frame.add(peelContainer, BorderLayout.CENTER);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setSize(800, 600);
		frame.setVisible(true);

	}

}
